import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.zip.*;
import java.util.ArrayList;
import javax.swing.*;

public class PGFViewer
{
	public static final byte MODE_8BIT_PALETTE =			0x00;
	public static final byte MODE_8BIT_GRAYSCALE =			0x01;
	public static final byte MODE_16BIT_COLOR =				0x02;
	public static final byte MODE_8BIT_PALETTE_ALPHA =		0x03;
	public static final byte MODE_8BIT_GRAYSCALE_ALPHA =	0x04;
	public static final byte MODE_24BIT_COLOR =				0x05;
	public static final byte MODE_16BIT_COLOR_ALPHA =		0x06;
	public static final byte MODE_32BIT =					0x07;

	public static void main(String[] args)
	{
		if(args.length != 1)
		{
			System.out.println("Missing file argument.");
		}
		Image img = null;
		try
		{
			img = readImage(args[0]);
		}
		catch(Exception e)
		{
			e.printStackTrace();
			System.exit(1);
		}
		DrawingCanvas canvas = new DrawingCanvas(img);
		JFrame frame = new JFrame("PGF Viewer - " + args[0]);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.add(canvas);
		frame.pack();
		frame.setVisible(true);
	}
	public static Image readImage(String file) throws IOException
	{
		FileInputStream in = new FileInputStream(file);
		char[] id = new char[] {
			(char) in.read(), (char) in.read(), (char) in.read(), (char) in.read()
		};
		if(id[0] != 'P' || id[1] != 'G' || id[2] != 'F' || id[3] != 0x00)
		{
			System.out.println("Input is not a valid PGF.");
			System.exit(0);
		}
		byte[] header = new byte[5];
		in.read(header);
		byte mode = header[4];
		BufferedImage img = new BufferedImage(unsigned(header[0]) + (unsigned(header[1])<<8), unsigned(header[2]) + (unsigned(header[3])<<8), BufferedImage.TYPE_INT_ARGB);

		//Standards say default to gray scale palette
		Color[] palette = new Color[256];
		for(int i = 0;i < 256;i++)
			palette[i] = new Color(i, i, i);

		while(in.available() != 0)
		{
			byte[] chunkHeader = new byte[12];
			in.read(chunkHeader);
			long compressedSize = (long) unsigned(chunkHeader[4]) | ((long) unsigned(chunkHeader[5])<<8) | ((long) unsigned(chunkHeader[6])<<16) | ((long) unsigned(chunkHeader[7])<<24);
			long uncompressedSize = (long) unsigned(chunkHeader[8]) | ((long) unsigned(chunkHeader[9])<<8) | ((long) unsigned(chunkHeader[10])<<16) | ((long) unsigned(chunkHeader[11])<<24);
			byte[] cdata = new byte[(int) compressedSize];
			byte[] data = new byte[(int) uncompressedSize];
			in.read(cdata);
			try
			{
				if(compressedSize != uncompressedSize)
				{
					Inflater inf = new Inflater();
					inf.setInput(cdata);
					inf.inflate(data);
					inf.end();
				}
				else
				{
					for(int i = 0;i < compressedSize;i++)
						data[i] = cdata[i];
				}
			}
			catch(Exception e)
			{
				e.printStackTrace();
				System.exit(1);
			}

			if(chunkHeader[0] == 'P' && chunkHeader[1] == 'L' && chunkHeader[2] == 'T' && chunkHeader[3] == 'E')
			{
				short numColors = unsigned(data[0]);
				for(int i = 0;i < numColors;i++)
				{
					System.out.println(unsigned(data[(3*i)+1]) + " " + unsigned(data[(3*i)+2]) + " " + unsigned(data[(3*i)+3]));
					palette[i] = new Color(unsigned(data[(3*i)+1]), unsigned(data[(3*i)+2]), unsigned(data[(3*i)+3]));
				}
			}
			else if(chunkHeader[0] == 'I' && chunkHeader[1] == 'D' && chunkHeader[2] == 'A' && chunkHeader[3] == 'T')
			{
				long[] pointers = new long[img.getWidth()];
				for(int i = 0;i < img.getWidth();i++)
				{
					pointers[i] = unsigned(data[i*4]) + (unsigned(data[(i*4)+1])<<8) + (unsigned(data[(i*4)+2])<<16) + (unsigned(data[(i*4)+3])<<24);
					for(int pos = (int) pointers[i];unsigned(data[pos]) + (unsigned(data[pos+1])<<8) != 0xFFFF;)
					{
						int y = unsigned(data[pos]) + (unsigned(data[pos+1])<<8);
						int length = unsigned(data[pos+2]) + (unsigned(data[pos+3])<<8);
						pos += 4;
						for(int j = 0;j < length;j++)
						{
							switch(mode)
							{
								case MODE_32BIT:
									img.setRGB(i, y+j, unsigned(data[pos]) | (unsigned(data[pos+1])<<8) | (unsigned(data[pos+2])<<16) | (unsigned(data[pos+3])<<24));
									pos += 4;
									break;
								case MODE_24BIT_COLOR:
									img.setRGB(i, y+j, unsigned(data[pos]) | (unsigned(data[pos+1])<<8) | (unsigned(data[pos+2])<<16) | (0xFF000000));
									pos += 3;
									break;
								case MODE_16BIT_COLOR_ALPHA:
								{
									int r = (int) Math.round((double) (unsigned(data[pos])&0x1F) * (255.0/31.0));
									int g = (int) Math.round((double) (((unsigned(data[pos])>>5)|(unsigned(data[pos+1])<<3))&0x3F) * (255.0/63.0));
									int b = (int) Math.round((double) ((unsigned(data[pos+1])&0xF8)>>3) * (255.0/31.0));
									img.setRGB(i, y+j, r | (g<<8) | (b<<16) | (unsigned(data[pos+2])<<24));
									pos += 3;
									break;
								}
								case MODE_16BIT_COLOR:
								{
									int r = (int) Math.round((double) (unsigned(data[pos])&0x1F) * (255.0/31.0));
									int g = (int) Math.round((double) (((unsigned(data[pos])>>5)|(unsigned(data[pos+1])<<3))&0x3F) * (255.0/63.0));
									int b = (int) Math.round((double) ((unsigned(data[pos+1])&0xF8)>>3) * (255.0/31.0));
									img.setRGB(i, y+j, r | (g<<8) | (b<<16) | 0xFF000000);
									pos += 2;
									break;
								}
								case MODE_8BIT_GRAYSCALE_ALPHA:
									img.setRGB(i, y+j, unsigned(data[pos]) | (unsigned(data[pos])<<8) | (unsigned(data[pos])<<16) | (unsigned(data[pos+1])<<24));
									pos += 2;
									break;
								case MODE_8BIT_PALETTE_ALPHA:
									img.setRGB(i, y+j, (palette[unsigned(data[pos])].getRGB() & 0xFFFFFF) | (unsigned(data[pos+1])<<24));
									pos += 2;
									break;
								case MODE_8BIT_GRAYSCALE:
									img.setRGB(i, y+j, unsigned(data[pos]) | (unsigned(data[pos])<<8) | (unsigned(data[pos])<<16) | (0xFF000000));
									pos++;
									break;
								default: //MODE_8BIT_PALETTE
									img.setRGB(i, y+j, palette[unsigned(data[pos])].getRed() | (palette[unsigned(data[pos])].getGreen()<<8) | (palette[unsigned(data[pos])].getBlue()<<16) | (0xFF000000));
									pos++;
									break;
							}
						}
					}
				}
			}
		}
		return img;
	}

	public static short unsigned(byte x)
	{
		if(x < 0)
			return (short) (256 + (short) x);
		return x;
	}
}

class DrawingCanvas extends Canvas
{
	Image img = null;

	public DrawingCanvas(Image img)
	{
		this.img = img;
		setSize(new Dimension(img.getWidth(this), img.getHeight(this)));
	}

	public void paint(Graphics g)
	{
		g.drawImage(img, 0, 0, this);
	}
}
